a, a:visited, a:focus {
  color: rgb(240, 110, 163);
  text-decoration: none;
}

a:hover, a:active {
  color: rgb(204, 94, 139);
  text-decoration: none;
}

.wy-menu-vertical header, .wy-menu-vertical p.caption {
    color: white;
    font-weight: 600;
}

.wy-menu-vertical a {
    color: rgb(200, 178, 188);
}

.wy-menu-vertical a:hover {
    background: rgb(240, 110, 163);
    color: white;
}

.wy-menu-vertical header, .wy-menu-vertical p.caption {
    color: black;
    font-weight: 600;
}

.wy-menu-vertical a{
    color: black;
}
.wy-nav-content-wrap, .wy-menu li.current > a {
    background-color: white;
}

.wy-nav-side {
    background: rgb(255, 200, 225);
}

.wy-nav-top {
    background: rgb(250, 231, 242);
}

.wy-nav-top a {
    color: black;
}

.wy-nav-top i {
    color: rgb(176, 46, 118);
}

.wy-side-nav-search {
    background: rgb(240, 110, 163);
}

.wy-side-nav-search > div.version {
    color: white;
}

.wy-side-nav-search > a {
    color: white;
}

pre {
    background: rgb(255, 230, 240);
}

html.writer-html4 .rst-content dl:not(.docutils) > dt,
html.writer-html5 .rst-content dl[class]:not(.option-list):not(.field-list):not(.footnote):not(.citation):not(.glossary):not(.simple) > dt {
    background: rgb(255, 230, 240);
    border-top: 3px solid #de6a98;
    color: rgb(240, 110, 163);
}

.rst-content .note {
    background: rgb(255, 230, 240);
}

.rst-content .note .admonition-title {
    background: rgb(240, 110, 163);
}

.rst-content code.literal {
    color: rgb(220, 40, 115);
}
